@@ -16,7 +16,6 @@ function(n, method=c("mean","median","HL1","HL2","HL3","sd","range","mad","shamo
   method = match.arg(method)
   if ( method %in% c("mean","sd", "range") ) return(0.0)
 
-
   switch (method,
      HL1 = { tmp = floor( (n*(n-1)-2)/4 )
                m = floor( ((2*n-1) - sqrt((2*n-1)^2-8*tmp))/2 ) },
@@ -48,7 +47,6 @@ HL <- function(x, method=c("HL1","HL2","HL3"),na.rm=FALSE) {
       HL2 = 0.5*median(xx[lower.tri(xx,diag=TRUE)]),
       HL3 = 0.5*median(xx) 
    )
-
    return( HL.estimation )
 }
 #============================================
@@ -97,7 +95,7 @@ c4.factor = function(n, method=c("sd","range","mad","shamos")) {
 mad.unbiased = function(x,center=median(x),constant=1.4826,na.rm=FALSE){
   if (na.rm) x = x[!is.na(x)]
   n = length(x)
-  if (n == 1) return(0)
+  if (n == 1L) return(0)
   constant*median(abs(x-center))/c4.factor(n,"mad") 
 }
 #-----
@@ -105,7 +103,7 @@ mad.unbiased = function(x,center=median(x),constant=1.4826,na.rm=FALSE){
 shamos.unbiased = function(x,constant=1.048358,na.rm=FALSE,IncludeEqual=FALSE){
   if (na.rm) x = x[!is.na(x)]
   n = length(x)
-  if (n == 1) return(0)
+  if (n == 1L) return(0)
   w1 = outer(x,x, "-")
   w2 = abs( w1[lower.tri(w1,diag=IncludeEqual)]  )
   constant * median( w2 ) / c4.factor(n,"shamos") 
@@ -132,18 +130,10 @@ function(n, method=c("mean", "median","HL1","HL2","HL3",
           mad = return(NA),
        shamos = return(NA) )
   }
-
-     mu = c4.factor(n, "range")
+  mu = c4.factor(n, "range")
 #-----------------------------------
-  d3 = function(n) {
-       f= function(x,y) {
-          (y^2) * exp(-(x^2+(x+y)^2)/2) * (pnorm(x+y)-pnorm(x))^(n-2)
-       }
-       tmp = integrate(function(y) {
-             sapply(y,function(y) {
-                  integrate(function(x){f(x,y)},-Inf,Inf)$value})}, 0, Inf)
-       sqrt( n*(n-1)/2/pi*tmp$value - c4.factor(n, "range")^2 )
-  }
+  d3 = factors.cc(n,factor="d3")
+  d3sq = d3^2 
   #------------------
   if ( n <= 100L ) {
     re = switch(method,
@@ -151,7 +141,7 @@ function(n, method=c("mean", "median","HL1","HL2","HL3",
            HL1    = (1/n.times.eVar.of.HL1[n]), 
            HL2    = (1/n.times.eVar.of.HL2[n]), 
            HL3    = (1/n.times.eVar.of.HL3[n]), 
-           range  = ((1-c4.factor(n,"sd")^2) / d3(n)^2 ), 
+           range  = ((1-c4.factor(n,"sd")^2) / d3sq ), 
            mad    = ((1-c4.factor(n,"sd")^2)/(n.times.eVar.of.mad[n]/n)),
            shamos = ((1-c4.factor(n,"sd")^2)/(n.times.eVar.of.shamos[n]/n)) )
   } else {
@@ -161,7 +151,7 @@ function(n, method=c("mean", "median","HL1","HL2","HL3",
            HL1    = (1/(1.0472 + 0.1127/n + 0.8365/n^2)), 
            HL2    = (1/(1.0472 + 0.2923/n + 0.2258/n^2)), 
            HL3    = (1/(1.0472 + 0.2022/n + 0.4343/n^2)), 
-           range  = ((1-c4.factor(n,"sd")^2) / d3(n)^2 ), 
+           range  = ((1-c4.factor(n,"sd")^2) / d3sq ), 
            mad    = {tmp=ifelse(n%%2L==1L,0.2996-149.357/n,-2.417-153.01/n);
                      (1/(2.7027+tmp/n))},
            shamos = (1/(1.15875+2.822/n+12.238/n^2)) )
@@ -218,7 +208,7 @@ w4.factor =  function(n, method=c("mad2","shamos2")) {
 mad2.unbiased = function(x,center=median(x),constant=1.4826,na.rm=FALSE){
   if (na.rm) x = x[!is.na(x)]
   n = length(x)
-  if (n == 1) return(0)
+  if (n == 1L) return(0)
   mad(x, center, constant, na.rm)^2 / w4.factor(n,"mad2") 
 }
 #-----
@@ -226,23 +216,57 @@ mad2.unbiased = function(x,center=median(x),constant=1.4826,na.rm=FALSE){
 shamos2.unbiased = function(x,constant=1.048358,na.rm=FALSE,IncludeEqual=FALSE){
   if (na.rm) x = x[!is.na(x)]
   n = length(x)
-  if (n == 1) return(0)
+  if (n == 1L) return(0)
 
   shamos(x, constant, na.rm, IncludeEqual)^2 / w4.factor(n,"shamos2")
 }
 #============================================
-# The American Standard uses "3*sigma"
-# The British Standard  uses "3.09*sigma"
-# Control Limits Factor for the sample standard deviation (s). 
-A3.factor = function(n, method=c("sd","range","mad","shamos"), sigma.factor=3){
+# Factors for constructing control charts 
+factors.cc = function(n, 
+  factor=c("A", "A1","A2","A3","B1","B2","B3","B4","B5","B6", 
+           "c2","c4","d2","d3","D1","D2", "D3","D4", "E1","E2","E3"), sigma.factor=3) 
+{
   n = as.integer(n)
   if ( n <= 1L ) return(NA)
-  method = match.arg(method)
-  switch(method,
-         sd = return(sigma.factor/c4.factor(n,"sd")/sqrt(n)),
-      range = return(sigma.factor/c4.factor(n,"range")/sqrt(n)),
-        mad = return(sigma.factor/c4.factor(n,"mad")/sqrt(n)),
-     shamos = return(sigma.factor/c4.factor(n,"shamos")/sqrt(n)) )
+  factor = match.arg(factor) 
+
+  c2fn = function(n) {sqrt(2/(n  )) * exp(lgamma(n/2) - lgamma((n-1)/2))}
+  c4fn = function(n) {sqrt(2/(n-1)) * exp(lgamma(n/2) - lgamma((n-1)/2))}
+  d3fn = function(n) {
+        d2n = c4.factor(n,method="range")
+        mu = d2n
+        joint2 = function(x,y) {(y-mu)^2*exp(-(x^2+(x+y)^2)/2)*(pnorm(x+y)-pnorm(x))^(n-2)}
+        tmp = integrate(function(y) {
+        sapply(y,function(y) {integrate(function(x) joint2(x,y),-Inf,Inf)$value})
+        }, 0, Inf)
+        sqrt( n*(n-1)/2/pi*tmp$value )
+  }
+
+  res =  switch (factor,
+      A  = {sigma.factor/sqrt(n)}, 
+      A1 = {sigma.factor/c2fn(n)/sqrt(n)}, 
+      A2 = {d2n=c4.factor(n,method="range"); sigma.factor/d2n/sqrt(n)}, 
+      A3 = {sigma.factor/c4fn(n)/sqrt(n)},    
+      c2 = {c2fn(n)},
+      c4 = {c4fn(n)},
+      B1 = {c2n=c2fn(n); max(0,c2n-sigma.factor*sqrt((n-1)/n-c2n^2))},
+      B2 = {c2n=c2fn(n); c2n+sigma.factor*sqrt((n-1)/n-c2n^2)},
+      B3 = {c4n=c4fn(n); max(0,1-sigma.factor/c4n*sqrt(1-c4n^2))},
+      B4 = {c4n=c4fn(n);       1+sigma.factor/c4n*sqrt(1-c4n^2)},
+      B5 = {c4n=c4fn(n); max(0,c4n-sigma.factor*sqrt(1-c4n^2))},
+      B6 = {c4n=c4fn(n);       c4n+sigma.factor*sqrt(1-c4n^2)}, 
+      d2 = c4.factor(n,method="range"),
+      d3 = d3fn(n),
+      D1 = {d2n=c4.factor(n,method="range"); max(0,d2n-sigma.factor*d3fn(n))},  
+      D2 = {d2n=c4.factor(n,method="range"); d2n+sigma.factor*d3fn(n)},
+      D3 = {d2n=c4.factor(n,method="range"); max(0,1-sigma.factor*d3fn(n)/d2n)}, 
+      D4 = {d2n=c4.factor(n,method="range"); 1+sigma.factor*d3fn(n)/d2n},
+      E1 = {sigma.factor/c2fn(n)}, 
+      E2 = {d2n=c4.factor(n,method="range"); sigma.factor/d2n},
+      E3 = {sigma.factor/c4fn(n)}
+  )
+  names(res) = factor
+  return(res)
 }
 #============================================
 
